package com.sun.jndi.ldap;

import java.lang.reflect.Proxy;
import java.io.ObjectStreamClass;
import java.io.InputStream;
import javax.naming.spi.DirStateFactory;
import javax.naming.directory.BasicAttributes;
import javax.naming.spi.DirectoryManager;
import java.util.Hashtable;
import javax.naming.Context;
import javax.naming.Name;
import java.io.ObjectInputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import javax.naming.NamingEnumeration;
import sun.misc.BASE64Decoder;
import javax.naming.directory.InvalidAttributeValueException;
import javax.naming.directory.InvalidAttributesException;
import javax.naming.RefAddr;
import sun.misc.BASE64Encoder;
import javax.naming.StringRefAddr;
import java.io.IOException;
import java.util.Vector;
import java.util.StringTokenizer;
import javax.naming.NamingException;
import javax.naming.directory.BasicAttribute;
import javax.naming.directory.DirContext;
import java.io.Serializable;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.directory.Attribute;
import javax.naming.directory.Attributes;

final class Obj
{
    static VersionHelper helper;
    static final String[] JAVA_ATTRIBUTES;
    static final int OBJECT_CLASS = 0;
    static final int SERIALIZED_DATA = 1;
    static final int CLASSNAME = 2;
    static final int FACTORY = 3;
    static final int CODEBASE = 4;
    static final int REF_ADDR = 5;
    static final int TYPENAME = 6;
    @Deprecated
    private static final int REMOTE_LOC = 7;
    static final String[] JAVA_OBJECT_CLASSES;
    static final String[] JAVA_OBJECT_CLASSES_LOWER;
    static final int STRUCTURAL = 0;
    static final int BASE_OBJECT = 1;
    static final int REF_OBJECT = 2;
    static final int SER_OBJECT = 3;
    static final int MAR_OBJECT = 4;
    
    private Obj() {
    }
    
    private static Attributes encodeObject(char c, Object o, Attributes attributes, Attribute attribute, boolean b) throws NamingException {
        if (attribute.size() == 0 || (attribute.size() == 1 && attribute.contains("top"))) {
            attribute.add(Obj.JAVA_OBJECT_CLASSES[0]);
        }
        if ((o instanceof Referenceable)) {
            attribute.add(Obj.JAVA_OBJECT_CLASSES[1]);
            attribute.add(Obj.JAVA_OBJECT_CLASSES[2]);
            if (!b) {
                attributes = (Attributes)attributes.clone();
            }
            attributes.put(attribute);
            return encodeReference(c, ((Referenceable)o).getReference(), attributes, o);
        }
        else if ((o instanceof Reference)) {
            attribute.add(Obj.JAVA_OBJECT_CLASSES[1]);
            attribute.add(Obj.JAVA_OBJECT_CLASSES[2]);
            if (!b) {
                attributes = (Attributes)attributes.clone();
            }
            attributes.put(attribute);
            return encodeReference(c, (Reference)o, attributes, null);
        }
        else {
            if ((o instanceof Serializable)) {
                attribute.add(Obj.JAVA_OBJECT_CLASSES[1]);
                if (!attribute.contains(Obj.JAVA_OBJECT_CLASSES[4]) && !attribute.contains(Obj.JAVA_OBJECT_CLASSES_LOWER[4])) {
                    attribute.add(Obj.JAVA_OBJECT_CLASSES[3]);
                }
                if (!b) {
                    attributes = (Attributes)attributes.clone();
                }
                attributes.put(attribute);
                attributes.put(new BasicAttribute(Obj.JAVA_ATTRIBUTES[1], serializeObject(o)));
                if (attributes.get(Obj.JAVA_ATTRIBUTES[2]) == null) {
                    attributes.put(Obj.JAVA_ATTRIBUTES[2], o.getClass().getName());
                }
                if (attributes.get(Obj.JAVA_ATTRIBUTES[6]) == null) {
                    Attribute typeNameAttr = LdapCtxFactory.createTypeNameAttr(o.getClass());
                    if (typeNameAttr != null) {
                        attributes.put(typeNameAttr);
                    }
                }
            }
            else if (!(o instanceof DirContext)) {
                throw new IllegalArgumentException("can only bind Referenceable, Serializable, DirContext");
            }
            return attributes;
        }
    }
    
    private static String[] getCodebases(Attribute attribute) throws NamingException {
        if (attribute == null) {
            return null;
        }
        StringTokenizer stringTokenizer = new StringTokenizer((String)attribute.get());
        Vector vector = new Vector(10);
        while (stringTokenizer.hasMoreTokens()) {
            vector.addElement(stringTokenizer.nextToken());
        }
        String[] array = new String[vector.size()];
        for (int i = 0; i < array.length; ++i) {
            array[i] = (String)vector.elementAt(i);
        }
        return array;
    }
    
    static Object decodeObject(Attributes attributes) throws NamingException {
        String[] codebases = getCodebases(attributes.get(Obj.JAVA_ATTRIBUTES[4]));
        try {
            Attribute value;
            if ((value = attributes.get(Obj.JAVA_ATTRIBUTES[1])) != null) {
                return deserializeObject((byte[])(byte[])value.get(), Obj.helper.getURLClassLoader(codebases));
            }
            Attribute value2;
            if ((value2 = attributes.get(Obj.JAVA_ATTRIBUTES[7])) != null) {
                return decodeRmiObject((String)attributes.get(Obj.JAVA_ATTRIBUTES[2]).get(), (String)value2.get(), codebases);
            }
            Attribute value3 = attributes.get(Obj.JAVA_ATTRIBUTES[0]);
            if (value3 != null && (value3.contains(Obj.JAVA_OBJECT_CLASSES[2]) || value3.contains(Obj.JAVA_OBJECT_CLASSES_LOWER[2]))) {
                return decodeReference(attributes, codebases);
            }
            return null;
        }
        catch (IOException rootCause) {
            NamingException ex = new NamingException();
            ex.setRootCause(rootCause);
            throw ex;
        }
    }
    
    private static Attributes encodeReference(char c, Reference reference, Attributes attributes, Object o) throws NamingException {
        if (reference == null) {
            return attributes;
        }
        String className;
        if ((className = reference.getClassName()) != null) {
            attributes.put(new BasicAttribute(Obj.JAVA_ATTRIBUTES[2], className));
        }
        String factoryClassName;
        if ((factoryClassName = reference.getFactoryClassName()) != null) {
            attributes.put(new BasicAttribute(Obj.JAVA_ATTRIBUTES[3], factoryClassName));
        }
        String factoryClassLocation;
        if ((factoryClassLocation = reference.getFactoryClassLocation()) != null) {
            attributes.put(new BasicAttribute(Obj.JAVA_ATTRIBUTES[4], factoryClassLocation));
        }
        if (o != null && attributes.get(Obj.JAVA_ATTRIBUTES[6]) != null) {
            Attribute typeNameAttr = LdapCtxFactory.createTypeNameAttr(o.getClass());
            if (typeNameAttr != null) {
                attributes.put(typeNameAttr);
            }
        }
        int size = reference.size();
        if (size > 0) {
            BasicAttribute basicAttribute = new BasicAttribute(Obj.JAVA_ATTRIBUTES[5]);
            Object o2 = null;
            for (int i = 0; i < size; ++i) {
                RefAddr value = reference.get(i);
                if ((value instanceof StringRefAddr)) {
                    basicAttribute.add("" + c + i + c + value.getType() + c + value.getContent());
                }
                else {
                    if (o2 == null) {
                        o2 = new BASE64Encoder();
                    }
                    basicAttribute.add("" + c + i + c + value.getType() + c + c + o2.encodeBuffer(serializeObject(value)));
                }
            }
            attributes.put(basicAttribute);
            return attributes;
        }
        return attributes;
    }
    
    private static Object decodeRmiObject(String className, String addr, String[] array) throws NamingException {
        return new Reference(className, new StringRefAddr("URL", addr));
    }
    
    private static Reference decodeReference(Attributes attributes, String[] array) throws NamingException, IOException {
        String factory = null;
        Attribute value;
        if ((value = attributes.get(Obj.JAVA_ATTRIBUTES[2])) != null) {
            String className = (String)value.get();
            Attribute value2;
            if ((value2 = attributes.get(Obj.JAVA_ATTRIBUTES[3])) != null) {
                factory = (String)value2.get();
            }
            Reference o = new Reference(className, factory, (array != null) ? array[0] : null);
            Attribute value3;
            if ((value3 = attributes.get(Obj.JAVA_ATTRIBUTES[5])) != null) {
                Object o2 = null;
                ClassLoader urlClassLoader = Obj.helper.getURLClassLoader(array);
                Vector o3 = new Vector();
                o3.setSize(value3.size());
                NamingEnumeration<Object> all = value3.getAll();
                while (all.hasMore()) {
                    String s = (String)all.next();
                    if (s.length() == 0) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - empty attribute value");
                    }
                    int char1 = s.charAt(0);
                    int n = 1;
                    int index;
                    if ((index = s.indexOf(char1, n)) < 0) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - separator '" + (char)char1 + "'not found");
                    }
                    String substring;
                    if ((substring = s.substring(n, index)) == null) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - empty RefAddr position");
                    }
                    int int1;
                    try {
                        int1 = Integer.parseInt(substring);
                    }
                    catch (NumberFormatException ex) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - RefAddr position not an integer");
                    }
                    int n2 = index + 1;
                    int index2;
                    if ((index2 = s.indexOf(char1, n2)) < 0) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - RefAddr type not found");
                    }
                    String substring2;
                    if ((substring2 = s.substring(n2, index2)) == null) {
                        throw new InvalidAttributeValueException("malformed " + Obj.JAVA_ATTRIBUTES[5] + " attribute - empty RefAddr type");
                    }
                    int beginIndex = index2 + 1;
                    if (beginIndex == s.length()) {
                        o3.setElementAt(new StringRefAddr(substring2, null), int1);
                    }
                    else if (s.charAt(beginIndex) == char1) {
                        ++beginIndex;
                        if (o2 == null) {
                            o2 = new BASE64Decoder();
                        }
                        o3.setElementAt((Object)deserializeObject(o2.decodeBuffer(s.substring(beginIndex)), urlClassLoader), int1);
                    }
                    else {
                        o3.setElementAt(new StringRefAddr(substring2, s.substring(beginIndex)), int1);
                    }
                }
                for (int i = 0; i < o3.size(); ++i) {
                    o.add((RefAddr)o3.elementAt(i));
                }
                return o;
            }
            return o;
        }
        else {
            throw new InvalidAttributesException(Obj.JAVA_ATTRIBUTES[2] + " attribute is required");
        }
    }
    
    private static byte[] serializeObject(Object obj) throws NamingException {
        try {
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            try (final ObjectOutputStream objectOutputStream = new ObjectOutputStream(out)) {
                objectOutputStream.writeObject(obj);
            }
            return out.toByteArray();
        }
        catch (IOException rootCause) {
            NamingException o = new NamingException();
            o.setRootCause(rootCause);
            throw o;
        }
    }
    
    private static Object deserializeObject(byte[] buf, ClassLoader classLoader) throws NamingException {
        try {
            ByteArrayInputStream in = new ByteArrayInputStream(buf);
            try (final ObjectInputStream objectInputStream = (classLoader == null) ? new ObjectInputStream(in) : new Obj.LoaderInputStream(in, classLoader)) {
                return objectInputStream.readObject();
            }
            catch (ClassNotFoundException rootCause) {
                NamingException o = new NamingException();
                o.setRootCause(rootCause);
                throw o;
            }
        }
        catch (IOException rootCause2) {
            NamingException o2 = new NamingException();
            o2.setRootCause(rootCause2);
            throw o2;
        }
    }
    
    static Attributes determineBindAttrs(char c, Object object, Attributes attrs, boolean b, Name name, Context nameCtx, Hashtable<?, ?> environment) throws NamingException {
        DirStateFactory.Result stateToBind = DirectoryManager.getStateToBind(object, name, nameCtx, environment, attrs);
        object = stateToBind.getObject();
        attrs = stateToBind.getAttributes();
        if (object == null) {
            return attrs;
        }
        if (attrs == null && object instanceof DirContext) {
            b = true;
            attrs = ((DirContext)object).getAttributes("");
        }
        boolean n = false;
        Object o;
        if (attrs == null || attrs.size() == 0) {
            attrs = new BasicAttributes(true);
            b = true;
            o = new BasicAttribute("objectClass", "top");
        }
        else {
            o = attrs.get("objectClass");
            if (o == null && !attrs.isCaseIgnored()) {
                o = attrs.get("objectclass");
            }
            if (o == null) {
                o = new BasicAttribute("objectClass", "top");
            }
            else if (n || !b) {
                o = (Object)o.clone();
            }
        }
        attrs = encodeObject(c, object, attrs, o, b);
        return attrs;
    }
    
    static {
        Obj.helper = VersionHelper.getVersionHelper();
        JAVA_ATTRIBUTES = new String[] { "objectClass", "javaSerializedData", "javaClassName", "javaFactory", "javaCodeBase", "javaReferenceAddress", "javaClassNames", "javaRemoteLocation" };
        JAVA_OBJECT_CLASSES = new String[] { "javaContainer", "javaObject", "javaNamingReference", "javaSerializedObject", "javaMarshalledObject" };
        JAVA_OBJECT_CLASSES_LOWER = new String[] { "javacontainer", "javaobject", "javanamingreference", "javaserializedobject", "javamarshalledobject" };
    }
    
    private static final class LoaderInputStream extends ObjectInputStream
    {
        private ClassLoader classLoader;
        
        LoaderInputStream(InputStream in, ClassLoader classLoader) throws IOException {
            super(in);
            this.classLoader = classLoader;
        }
        
        @Override
        protected Class<?> resolveClass(ObjectStreamClass desc) throws IOException, ClassNotFoundException {
            try {
                return this.classLoader.loadClass(desc.getName());
            }
            catch (ClassNotFoundException ex) {
                return super.resolveClass(desc);
            }
        }
        
        @Override
        protected Class<?> resolveProxyClass(String[] array) throws IOException, ClassNotFoundException {
            ClassLoader classLoader = null;
            int n = 0;
            Class[] interfaces = new Class[array.length];
            Class<Object> forName;
            for (int i = 0; i < array.length; interfaces[i] = forName, ++i) {
                forName = Class.forName(array[i], false, this.classLoader);
                if ((forName.getModifiers() & 0x1) == 0x0) {
                    if (n != 0) {
                        if (classLoader != forName.getClassLoader()) {
                            throw new IllegalAccessError("conflicting non-public interface class loaders");
                        }
                    }
                    else {
                        classLoader = forName.getClassLoader();
                        n = 1;
                    }
                }
            }
            try {
                return Proxy.getProxyClass((n != 0) ? classLoader : this.classLoader, interfaces);
            }
            catch (IllegalArgumentException ex) {
                throw new ClassNotFoundException(null, ex);
            }
        }
    }
}
